;; Generated package description from consult-hoogle.el  -*- no-byte-compile: t -*-
(define-package "consult-hoogle" "0.1.0" "Hoogle frontend using consult" '((emacs "27.1") (haskell-mode "16.1")) :commit "593094b154f0ae5a404724d6ac7b7e1592fac680" :keywords '("docs" "languages") :url "https://codeberg.org/rahguzar/consult-hoogle")
